/*
 * SPDX-FileCopyrightText: 2024...2025 Michael Margraf <dd6um@darc.de>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * ------------------------------------------------------------------
 * OpenAccess file reader
 * This program reads a file in OpenAccess (*.oa) format.
 * Then, it interprets the content and writes the result
 * into a text file (hex/ASCII, list of elements or Qucs schematic).
 *
 * Copyright (C) 2024...2025 Michael Margraf <dd6um@darc.de>
 *
 * This project is supported by the NLnet Foundation,
 * see https://nlnet.nl/project/foaHandler/
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ------------------------------------------------------------------
 * To build with MSVC compiler, type:
 * cl foaHandler.cpp /nologo
 *
 * To build with GCC compiler, type:
 * g++ *.cpp -Wall -o foaHandler.exe
 *
 * To build with Clang compiler, type:
 * clang++ *.cpp -Wall -o foaHandler.exe
 * ------------------------------------------------------------------ */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>

#define TEXT_SIZE  20  // number of pixel per text line

// ------------------------------------------------------------------------
// Function for printing a 64 bit integer as hexadecimal value and
// that works for all compilers (MSVC, MinGW, GCC, Clang, 32 bit, 64 bit).
char* strInt64(int64_t num)
{
  static char str[24];
  if(num & 0xFFFFFFFF00000000)
    sprintf(str, "0x%x%08x", int(num >> 32), (int)num);
  else
    sprintf(str, "0x%04x", (int)num);
  return str;
}

// ------------------------------------------------------------------------
// Converts a string to a number and also applies the prefix of the unit.
// E.g. "10 mA" returns 1e-2, "1.5pF" returns 1.5e-12 etc.
double str2num(char *p)
{
  double d = strtod(p, &p);

  while(*p == ' ')  // ignore spaces between number and unit
    p++;

  if(*p) {
    // check if number has a unit
    switch(*p) {
      case 'T': d *= 1e12;  break; // tera
      case 'G': d *= 1e9;   break; // giga
      case 'M': d *= 1e6;   break; // mega
      case 'k': d *= 1e3;   break; // kilo
      case 'd': d *= 1e-1;  break; // deci
      case 'c': d *= 1e-2;  break; // centi
      case 'm': d *= 1e-3;         // milli
        if(p[1] == 'i')
          if(p[2] == 'l')
            d *= 0.0254;           // mil = 1/1000 inch
        break;
      case 'u': d *= 1e-6;  break; // micro
      case 'n': d *= 1e-9;  break; // nano
      case 'p': d *= 1e-12; break; // pico
      case 'f': d *= 1e-15; break; // femto
    }
  }

  return d;
}

// ------------------------------------------------------------------------
// prints out the properties of a component
int outputTextFormat(FILE *outputFile,
                     char *name, char *lib, char *model, int property, int *position,
                     int *textPos, int numParams, char **paramValue, char **paramVisible)
{
  fprintf(outputFile,
      "  %s: %s.%s, enabled=%d, x=%d, y=%d, text=%d/%d, mirrored=%d, rotated=%d\n",
         name, lib, model,
         property & 3, // enabled, disabled, shorted
         position[0], position[1],  // component coordinates
         textPos[0], textPos[1],   // text coordinates
         property & 0x40 ? 1 : 0, // mirrored about y axis
         (property >> 4) & 3);   // n-times rotated
  fprintf(outputFile, "    parameters:");
  for(int i=0; i<numParams; i++)
    fprintf(outputFile, " \"%s\"%s", paramValue[i], paramVisible[i]);
  fprintf(outputFile, "\n");
  return 1;
}

// ------------------------------------------------------------------------
// prints out a component in Qucs 0.0.19 format
int outputQucsFormat(FILE *outputFile,
                     char *name, char *lib, char *model, int property, int *position,
                     int *textPos, int numParams, char **paramValue, char **paramVisible)
{
  // Note on coordinates:
  // In Keysight ADS a standard component (e.g. resistor) is 160 units long.
  // In Qucs all standard components (like capacitors) are 60 units long.
  // Therefore, all coordinates have to be divided by 8/3.
  // Furthermore, positive y-coordinates goes up, whereas
  // they go down in Qucs. Therefore, their sign is reversed here.
  int x = position[0] / +8 * 3;   // component coordinates
  int y = position[1] / -8 * 3;
  int tx = textPos[0]/+8 * 3 - x; // coordinates of component text
  int ty = textPos[1]/-8 * 3 - y - TEXT_SIZE;
  int enable = property & 3;        // enabled, disabled, shorted
  int rotate = property >> 4;       // rotated n-times (0...4)
  int mirror = (property >> 6) & 1; // mirrored about y axis
  if(mirror)
    rotate += 2; // in Qucs mirror is about x axis
  rotate &= 3;

  if(strcmp(model, "GROUND") == 0) { // ground component
    fprintf(outputFile, "  <GND * %d %d %d 0 0 0 %d>\n",
            enable, x, y, (rotate+1) & 3); // never mirrored!
    return 1;
  }

  if(strcmp(model, "VAR") == 0) { // variable component
    fprintf(outputFile, "  <Eqn %s %d %d %d %d %d 0 0",
            name, enable, x, y+10, tx-60, ty+26);
   for(int i=0; i<numParams; i+=2)
      fprintf(outputFile, " \"%s=%s\" 1", paramValue[i], paramValue[i+1]);
    fprintf(outputFile, " \"no\" 0>\n");
    return 1;
  }

  if(strcmp(model, "C") == 0) { // capacitor
    if(numParams < 1)
      return -1001;
    if(rotate & 1)
      y += 30;
    else
      x += 30;
    fprintf(outputFile, "  <C %s %d %d %d %d %d %d %d \"%s\" %s \"\" 0 \"neutral\" 0>\n",
            name, enable, x, y, tx, ty, mirror, rotate,
            paramValue[0], paramVisible[0]);
    return 1;
  }

  if(strcmp(model, "Term") == 0) { // s-parameter port
    if(numParams < 2)
      return - 1002;
    rotate = (rotate + 2) & 3;
    if(rotate & 1)
      y += 30;
    else
      x += 30;
    fprintf(outputFile, "  <Pac %s %d %d %d %d %d %d %d \"%s\" %s \"%s\" %s \"0 dBm\" 0 \"1 GHz\" 0 \"26.85\" 0>\n",
            name, enable, x, y, tx, ty, mirror, rotate,
            paramValue[0], paramVisible[0], paramValue[1], paramVisible[1]);
    return 1;
  }

  if(strcmp(model, "S2P") == 0) { // 2-port Touchstone file
    if(numParams < 1)
      return -1003;
    fprintf(outputFile, "  <SPfile %s %d %d %d %d %d %d %d %s %s \"rectangular\" 0 \"linear\" 0 \"open\" 0 \"2\" 0>\n",
            name, enable, x+30, y, tx, ty, mirror, rotate,
            paramValue[0], paramVisible[0]);
    return 1;
  }

  if(strcmp(model, "DC") == 0) { // DC simulation component
    fprintf(outputFile, "  <.DC %s %d %d %d %d %d 0 0 \"26.85\" 0 \"0.001\" 0 \"1 pA\" 0 \"1 uV\" 0 \"no\" 0 \"150\" 0 \"no\" 0 \"none\" 0 \"CroutLU\" 0>\n",
            name, enable, x, y+34, tx-10, ty-38);
    return 1;
  }

  if(strcmp(model, "S_Param") == 0) { // s-parameters simulation component
    if(numParams < 9)
      return -1004;
    char *start = paramValue[3];
    char *stop = paramValue[4];
    char *points = paramValue[8];
    fprintf(outputFile, "  <.SP %s %d %d %d %d %d 0 0 \"lin\" 0 \"%s\" 1 \"%s\" 1 \"%s\" 1 \"no\" 0 \"1\" 0 \"2\" 0>\n",
            name, enable, x, y+34, tx-10, ty-38, start, stop, points);
    return 1;
  }

  if(strcmp(model, "ParamSweep") == 0) { // parameter sweep component
    if(numParams < 14)
      return -1005;
    char *var = paramValue[2];  // sweep variable enclosed in ""
    if(*var == '"')
      var++;
    if(var[strlen(var)-1] == '"')
      var[strlen(var)-1] = 0;
    char *sim = paramValue[3];  // simulation instance enclosed in ""
    if(*sim == '"')
      sim++;
    if(sim[strlen(sim)-1] == '"')
      sim[strlen(sim)-1] = 0;
    char *start = paramValue[11];
    char *stop = paramValue[12];
    char *step = paramValue[13];
    int num = int((str2num(stop) - str2num(start)) / str2num(step) + 0.5) + 1;
    fprintf(outputFile, "  <.SW %s %d %d %d %d %d 0 0 \"%s\" 1 \"lin\" 0 \"%s\" 1 \"%s\" 1 \"%s\" 1 \"%d\" 1>\n",
            name, enable, x, y+34, tx-10, ty-38, sim, var, start, stop, num);
    return 1;
  }

  // Each standard component is outputed as a subcircuit component.
  fprintf(outputFile, "  <Sub %s %d %d %d %d %d %d %d \"foaLib_%s.sch\" 0",
          name, enable, x, y, tx, ty, mirror, rotate, model);
  for(int i=0; i<numParams; i++)
    fprintf(outputFile, " \"%s\" %s", paramValue[i], paramVisible[i]);
  fprintf(outputFile, ">\n");
  return 0;
}


// ------------------------------------------------------------------------
// This is a class for reading OpenAccess files. It's really open and free!
// ------------------------------------------------------------------------
class FoaHandler {
public:
  FoaHandler();
 ~FoaHandler();

  enum { outputFormatHex, outputFormatText, outputFormatQucs };

  int64_t loadFile(const char *filename);
  int checkFileFormat();
  int printFileContent();
  int printFileElements(int);

  int openOutputFile(const char*, const char*);
  void closeOutputFile();

private:
  int64_t fileSize;
  int32_t *fileContent;
  char *fileEnd;
  FILE *outputFile;

  static int printBlock(FILE*, unsigned char*, int64_t, int64_t, int64_t);
  static int printSubBlock(FILE*, unsigned char*, int64_t, int64_t, int64_t);

  char* getBlock(int64_t, int *num=0);
  char* getSubblock(char*, int64_t, int *num=0);
  int getValue(char*, int, int num=0, int row=0);
  int getWireCoordinate(char*&);
  int getComponentParameters(int, int*, char**&, char**&);
  int printSubcircuitPorts(int);
  int printComponents(int);
  int printWires(int);

  int stringStop, numObjects;
  char *stringList, *objectNameList;
  char* getString(char*, int);
  char* getObjectName(int);
};

// ------------------------------------------------------------------------
FoaHandler::FoaHandler()
{
  fileSize = -1;
  fileContent = 0;
  numObjects = 0;
  outputFile = stdout;
}

// ------------------------------------------------------------------------
FoaHandler::~FoaHandler()
{
  fclose(outputFile);
  free(fileContent);
}

// ------------------------------------------------------------------------
// This function creates the name of the output file and opens it.
// The filename is built from the input filename and a suffix.
// e.g. filename="sch.oa" and suffix=".txt" gives "sch.txt"
int FoaHandler::openOutputFile(const char *filename, const char *suffix)
{
  int len = strlen(filename);
  if(suffix)
    len += strlen(suffix);
  char *name = (char*)malloc((len+1) * sizeof(char));
  if(name == 0)
    return -1;
  strcpy(name, filename);

  if(suffix) {  // replace the suffix with a new one?
    char *p = strrchr(name, '.');
    if(p == 0)
      p = name + strlen(name);
    strcpy(p, suffix);  // replace old suffix with new one
  }

  printf("writing file: %s\n", name); fflush(stdout);
  FILE *file = fopen(name, "w");
  free(name);
  if(file == 0)
    return -2;

  outputFile = file;
  return 1;
}

// ------------------------------------------------------------------------
void FoaHandler::closeOutputFile()
{
  fclose(outputFile);
  outputFile = stdout;
}

// ------------------------------------------------------------------------
// This function loads the file content into memory.
// If successful, it returns the file size, otherwise it returns -1.
int64_t FoaHandler::loadFile(const char *filename)
{
  FILE *file;
  struct stat info;

  free(fileContent);
  fileContent = 0;
  fileSize = -1;

  if(stat(filename, &info) == 0) {
    file = fopen(filename, "rb");
    if(file) {
      fileContent = (int32_t*)malloc(info.st_size);
      if(fileContent) {
        if(fread(fileContent, 1, info.st_size, file) != (size_t)info.st_size) {
          free(fileContent);
          fileContent = 0;
        }
        else {
          fileSize = info.st_size;
          fileEnd = ((char*)fileContent) + fileSize - 4; // before file end flag
        }
      }
      fclose(file);
    }
  }

  return fileSize;
}

// ------------------------------------------------------------------------
// This function checks if the file content fits to an OpenAccess file.
// If so, it returns the number of main blocks, otherwise returns a negative number.
int FoaHandler::checkFileFormat()
{
  // first and last bytes of an OpenAccess file
  const int32_t OPENACCESS_FILE_ID = 0x01234567;

  if(fileSize < 0x0038) // memory for at least one block?
    return -10;
  if(fileContent[0] != OPENACCESS_FILE_ID) // Is format identification correct?
    return -11;
  if(fileContent[fileSize/sizeof(int32_t)-1] != OPENACCESS_FILE_ID)
    return -12;  // file end identification missing

  int32_t maxBlocks = fileContent[4]; // max number of blocks (size of block table)
  int32_t numBlocks = fileContent[5]; // number of used blocks
  if(numBlocks < 2) // main block (ID == 1) and end block (ID == 40) are mandatory
    return -13;
  if(maxBlocks < numBlocks)
    return -14;
  if(fileSize < (3*8*int64_t(maxBlocks) + 0x18)) // size of main block table plus file IDs
    return -15;

  return int(numBlocks);
}

// ------------------------------------------------------------------------
// This function handles a subblock of the file.
// It prints its content, and also prints its subsubblocks.
// Returns a negative number, if an error occured.
int FoaHandler::printSubBlock(FILE *file, unsigned char *content, int64_t pos, int64_t size, int64_t type)
{
  int32_t i, j, n, *p;
  content += pos;

  fprintf(file, "subblock %s (0x%04x bytes at 0x%04x):", strInt64(type), (int)size, (int)pos);
  if((pos < 0) || (size <= 0)) { // subblock empty?
    fprintf(file, "\n");
    return 0;
  }

  p = (int32_t*)content;
  if((size < 32) || (p[0] != 4)) { // no subblocks?
    for(i=0; i<size; i++) {
      if((i % 16) == 0)
        fprintf(file, "\n  %04x:", i);
      fprintf(file, " %02x", content[i]);
    }
    fprintf(file, "\n");
    return 1;
  }

  // This subblock contains further subblocks, so print them one by one.
  if(p[1] == 0)  // extra bytes?
    p++;
  pos = *(int64_t*)(p + 1);  // start of data after block table
  int32_t maxBlocks = p[3];  // max number of blocks (size of block table)
  int32_t numBlocks = p[4];
  p += 5;

  // output block header
  n = (unsigned char*)p - content;
  for(i=0; i<n; i++) {
    if((i % 16) == 0)
      fprintf(file, "\n  %04x:", i);
    fprintf(file, " %02x", content[i]);
  }
  fprintf(file, "\n");

  if(numBlocks > 0x10000) // avoid numerical overflow
    return -21;
  if(maxBlocks < numBlocks)
    return -22;
  n += 3*8*maxBlocks; // add size of block table
  if(size < n)       // block size too small?
    return -23;

  n -= 4;
  size -= 4;
  content += 4;
  int64_t len = 0;
  int64_t offset = 0;
  int64_t *blockID = (int64_t*)p;  // block table
  int64_t *blockPos = blockID + numBlocks;
  int64_t *blockSize = blockPos + numBlocks;
  for(i=0; i<numBlocks; i++) {
    if(blockPos[i] == 0) // belongs to previous (main) block?
      offset += pos + len;
    else if(blockPos[i] != (pos+len))  // end of main block?
      offset = 0;
    pos = blockPos[i];
    len = blockSize[i];
    
    if(len > 0)
      if(offset+pos < 0)
        return -24;
    if(offset+pos+len > size)
      return -25; // error: pointer outside of block data

    type = blockID[i];
    fprintf(file, "  subsubblock %s (%d bytes at 0x%x):", strInt64(type), int(len), int(offset+pos));
    for(j=0; j<len; j++)
      fprintf(file, " %02x", content[offset+pos+j]);
    fprintf(file, "\n");
  }

  return 2;
}

// ------------------------------------------------------------------------
// This function handles a main block of the file.
// It either prints its content, or if the block contains subblocks,
// it calls a print function for each of them.
// Returns a negative number, if an error occured.
int FoaHandler::printBlock(FILE *file, unsigned char *content, int64_t pos, int64_t size, int64_t type)
{
  int i, j, result;
  content += pos;

  fprintf(file, "**********************************************************************\n");
  fprintf(file, "block %s (0x%04x bytes at 0x%04x)", strInt64(type), (int)size, (int)pos);

  if((size < 24) || (*((int64_t*)content) != 4)) { // no subblocks?
    fprintf(file, " DATA");  // print as hex and ASCII block
    for(i=0; i<size; i++) {
      if((i % 16) == 0) {
        if(i >= 16)
          for(j=i-16; j<i; j++) {
            if((content[j] > ' ') && (content[j] < 'z'))
              fputc(content[j], file); // print data as ASCII
            else
              fputc('.', file);
          }
        fprintf(file, "\n%04x: ", i);
      }
      fprintf(file, "%02x ", content[i]); // print data as hex number
    }
    if((i % 16) == 0)
      i -= 16;
    else
      for(j=i%16; j<16; j++)
        fprintf(file, "   "); // align ASCII block
    for(i&=-16; i<size; i++) // print remaining ASCII
      if((content[i] > ' ') && (content[i] < 'z'))
        fputc(content[i], file);
      else
        fputc('.', file);
    fprintf(file, "\n");
    return 1;
  }


  // This block contains subblocks, so print them one by one.
  int32_t maxBlocks = *(int32_t*)(content + 16);  // size of block table
  int32_t numBlocks = *(int32_t*)(content + 20);
  fprintf(file, " with 0x%04x subblocks", numBlocks);
  pos = *(int64_t*)(content + 8);   // begin of block data

  // output block header
  for(i=0; i<24; i++) {
    if((i % 16) == 0)
      fprintf(file, "\n%04x:", i);
    fprintf(file, " %02x", content[i]);
  }
  fprintf(file, "\n\n");

  if(numBlocks > 0x10000) // avoid numerical overflow
    return -11;
  if(maxBlocks < numBlocks)
    return -12;
  int32_t n = 3*8*maxBlocks + 24; // size of block table plus header
  if(size < n)  // block size too small?
    return -13;

  // output subblocks
  n -= 4;
  size -= 4;
  content += 4;
  int64_t len = 0;
  int64_t offset = 0;
  int64_t *blockID = (int64_t*)(content + 20);
  int64_t *blockPos = blockID + numBlocks;
  int64_t *blockSize = blockPos + numBlocks;
  for(i=0; i<numBlocks; i++) {
    if(blockPos[i] == 0) // belongs to previous (main) block?
      offset += pos + len;
    else if(blockPos[i] != (pos+len))  // end of main block?
      offset = 0;
    pos = blockPos[i];
    len = blockSize[i];

    if(len > 0)
      if(offset+pos < 0)
        return -14; // error: pointer outside of block data
    if(offset+pos+len > size)
      return -15; // error: pointer outside of file data
    result = printSubBlock(file, content+offset, pos, len, blockID[i]);
    if(result < 0)
      return result;
  }

  return 2;
}

// ------------------------------------------------------------------------
// This function writes the decoded file content into the output file.
// It calls a print function for every block in the main table.
// Returns a negative number, if an error occured.
// Should be called only if checkFileFormat() was successful!
int FoaHandler::printFileContent()
{
  int32_t numBlocks = fileContent[5]; // number of main blocks
  fprintf(outputFile, "file size: %d Bytes\n", (int)fileSize);
  fprintf(outputFile, "number of blocks: %d\n\n", numBlocks);

  int result = -1;
  int64_t pos = 0;
  int64_t len = 0;
  int64_t offset = 0;
  int64_t *blockID = (int64_t*)(fileContent + 6);
  int64_t *blockPos = blockID + numBlocks;
  int64_t *blockSize = blockPos + numBlocks;

  printf("%d blocks: ", numBlocks); fflush(stdout);
  for(int i=0; i<numBlocks; i++) {  // going through all blocks
    printf("*"); fflush(stdout); // progressbar

    if(blockPos[i] < 0) {  // empty (dummy) block?
      fprintf(outputFile, "**********************************************************************\n");
      fprintf(outputFile, "empty block %s: %d, %d\n", strInt64(blockID[i]), (int)blockPos[i], (int)blockSize[i]);
      continue;
    }
    if(blockPos[i] == 0) // embedded in main block?
      offset += pos;
    else if(blockPos[i] != (pos+len)) // end of main block?
      offset = 0;
    pos = blockPos[i];
    len = blockSize[i];

    if(len > 0) {
      if(pos < 0) {
        result = -2;
        break;
      }
      if(offset+pos < 24) {
        result = -3;
        break;
      }
    }
    if((offset+pos+len) > fileSize) {
      result = -4;  // error: pointer outside of file data
      break;
    }
    result = printBlock(outputFile, (unsigned char*)fileContent, pos+offset, len, blockID[i]);
    if(result < 0)
      break;
  }

  printf("\n");
  return result;
}

// ------------------------------------------------------------------------
// Searches in the block table for the subblock with the ID given in "type".
// Returns a pointer to that block or zero if it wasn't found.
// If "num" isn't null, set it to the number of block objects (subblock 2).
char* FoaHandler::getSubblock(char *content, int64_t type, int *num)
{
  if(content+24 > fileEnd) // block table outside of file data?
    return 0;

  int32_t *p = (int32_t*)content;
  if(p[0] != 4)  // no subblocks?
    return 0;

  if(p[1] == 0)  // extra bytes?
    p++;
  int64_t pos = *(int64_t*)(p + 1);  // start of data after block table
  int32_t maxBlocks = p[3];  // max number of blocks (size of block table)
  int32_t numBlocks = p[4];
  p += 5;

  if(numBlocks > 0x10000) // avoid numerical overflow
    return 0;
  if(maxBlocks < numBlocks)
    return 0;
  if(((char*)p)+3*8*maxBlocks > fileEnd) // block table outside of file data?
    return 0;

  int64_t size = 0;
  int64_t offset = 0;
  int64_t *blockID = (int64_t*)p;
  int64_t *blockPos = blockID + numBlocks;
  int64_t *blockSize = blockPos + numBlocks;

  for(int32_t i=0; i<numBlocks; i++) { // search for block in table
    if(blockPos[i] == 0) // belongs to previous (main) block?
      offset += pos + size;
    else if(blockPos[i] != (pos+size))  // end of main block?
      offset = 0;
    pos = blockPos[i];
    size = blockSize[i];
    if(blockID[i] != type)  // Is it not the wanted block?
      continue;

    offset += pos + 4;
    if(offset < 0)
      return 0;
    content += offset;
    if(content+size > fileEnd) // block outside of file data?
      return 0;

    if(num) {
      p = (int32_t*)getSubblock(content, 2); // number of objects
      if(p == 0)
        *num = -1;
      else
        *num = *p;
    }
    return content;
  }

  return 0;  // block not found
}

// ------------------------------------------------------------------------
// Searches in the main table for the block with the ID given in "type".
// Returns a pointer to that block or zero if it wasn't found.
// If "num" isn't null, set it to the number of block objects (subblock 2).
// Should be called only if checkFileFormat() was successful!
char* FoaHandler::getBlock(int64_t type, int *num)
{
  int64_t size = 0;
  int64_t offset = 0;
  char *content = (char*)fileContent;
  int64_t pos = *(int64_t*)(content + 8);  // begin of block data
  int32_t numBlocks = *(int32_t*)(content + 0x0014);
  int64_t *blockID = (int64_t*)(content + 0x0018);
  int64_t *blockPos = blockID + numBlocks;
  int64_t *blockSize = blockPos + numBlocks;

  for(int32_t i=0; i<numBlocks; i++) { // search for block in table
    if(blockPos[i] < 0)  // empty (dummy) block?
      continue;
    if(blockPos[i] == 0) // embedded in main block?
      offset += pos;
    else if(blockPos[i] != (pos+size)) // end of main block?
      offset = 0;
    pos = blockPos[i];
    size = blockSize[i];
    if(blockID[i] != type)  // Is it not the wanted block?
      continue;

    offset += pos;
    if(offset < 24) // block outside of file data?
      return 0;
    if(offset+size > fileSize) // block outside of file data?
      return 0;

    content += offset;
    if(num) {
      int32_t *p = (int32_t*)getSubblock(content, 2); // number of objects
      if(p == 0)
        *num = -1;  // number not available
      else
        *num = *p;
    }
    return content;
  }

  return 0;  // block not found
}

// ------------------------------------------------------------------------
// Returns a value from the value list of a block.
// "content" is the pointer to the block, "index" the position in the value list.
// In matrix blocks, the number of values per row and
// the wanted row needs to be speficied, too.
int FoaHandler::getValue(char *content, int index, int num, int row)
{
  int32_t i, n, val, *pi;
  if(index < 0)
    return -1;

  pi = (int32_t*)content;
  if(pi+row+2 >= (int32_t*)fileEnd) // outside of file data?
    return -2;

  if(pi[0] > 0) { // a matrix block?
    i = pi[1];   // number of data rows
    if((row >= i) || (i < 1) || (i > 256))
      return -3;
    n = pi[row + 2]; // bytes per value
  }
  else if(pi[0] == (int32_t)0xC0000001) {
    n = pi[2]; // bytes per value
    i = 1;
  }
  else { // usually pi[0] == 0xC8000000
    n = pi[1]; // bytes per value
    i = 0;
  }
  if((n < 1) || (n > 8))
    return -4;

  content += (i + 2)*sizeof(int32_t) + n*(num*row + index); // position of data
  if(content >= fileEnd) // outside of file data?
    return -5;

  val = *(int32_t*)content;
  if(n == 1)
    val &= 0xFF;
  else
    val &= 0xFFFF;
  return val;
}

// ------------------------------------------------------------------------
// Reads the wire coordinate from position "p", decodes it
// and sets the pointer "p" to the next wire coordinate.
int FoaHandler::getWireCoordinate(char* &p)
{
  int val = *((unsigned int*)p);  // coordinate

  switch(val & 6) {
    case 2:
      val &= 0xFFFF;
      p += 2;
      break;
    case 4:
      val &= 0xFFFFFF;
      p += 3;
      break;
    case 6:
      val &= 0xFFFFFFFF;
      p += 4;
      break;
    default:
      val &= 0xFF;
      p++;
  }

  int pos = val >> 5;
  if(val & 1)
    pos *= 100;
  if(val & 0x10)
    pos *= -1;

  return pos;
}

// ------------------------------------------------------------------------
// Returns the string from block 0x000A with index "idx".
char* FoaHandler::getString(char *list, int idx)
{
  if(idx < 1)
    return 0;
  int pos = getValue(list, idx-1);
  if((pos < 1) || (pos > stringStop))
    return 0;
  return stringList + pos;
}

// ------------------------------------------------------------------------
// Returns the name of the object from block 0x000B subblock 0x0101 with index "idx".
char* FoaHandler::getObjectName(int idx)
{
  if((idx < 0) || (idx >= numObjects))
    return 0;
  int pos = getValue(objectNameList, idx) / 2;
  if((pos < 1) || (pos > stringStop))
    return 0;
  return stringList + pos;
}

// ------------------------------------------------------------------------
// This function writes all elements found in the file content into the output file.
// Returns a negative number, if an error occured.
// Parameter "format": see
// Should be called only if checkFileFormat() was successful!
int FoaHandler::printFileElements(int format)
{
  char *p;

  // get list of all ASCIIZ strings used in the file
  stringList = getBlock(0x000A);
  if(stringList == 0)
    return -10;
  stringStop = ((int32_t*)stringList)[1];
  stringList += 0x0013; // begin of ASCIIZ strings
  if(stringList+stringStop >= fileEnd) // outside of file data?
    return -11;
  if(stringList[stringStop] != 0) // string must be terminated here
    return -12;

  if(format == outputFormatText) {
    // print versions
    p = getBlock(0x001C); // OpenAccess compatibility block
    if(p)
      fprintf(outputFile, "OpenAccess compatibility: %s\n", p+2);
    p = getBlock(0x001D);  // OpenAccess version block
    if(p)
      fprintf(outputFile, "OpenAccess version: %s (%s)\n", p+0x18, p+0x28);
  }
  else
    fprintf(outputFile,
      "<Qucs Schematic 0.0.19>\n"
      "<Properties>\n");

  // get all object names
  p = getBlock(0x000B, &numObjects);
  if(p == 0)
    return -13;
  objectNameList = getSubblock(p, 0x0101);
  if(objectNameList == 0)
    return -14;

  if(format == outputFormatText) {
    // get cell name
    p = getBlock(0x010C);
    if(p) {
      p = getSubblock(p, 0x0300);
      if(p)
        fprintf(outputFile, "Cell name: %s\n", getObjectName(getValue(p, 0) - 1));
    }
  }

  // get coordinates of used area
  p = getBlock(0x0132);
  if(p) {
    int *areaXY = (int*)getSubblock(p, 0x0302);
    if(areaXY == 0)
      return -15;
    if(areaXY+4 >= (int32_t*)fileEnd)
      return -16;
    if(format == outputFormatText)
      fprintf(outputFile, "Area used by elements: %d/%d ... %d/%d\n",
                          areaXY[1], areaXY[2], areaXY[3], areaXY[4]); // x1, y1, x2, y2
    else
      fprintf(outputFile, "  <View=%d,%d,%d,%d,1,0,0>\n",
                             areaXY[1], areaXY[2], areaXY[3], areaXY[4]);
  }

  if(format != outputFormatText)
    fprintf(outputFile,
      "</Properties>\n"
      "<Symbol>\n"
      "</Symbol>\n"
      "<Components>\n");
  fflush(outputFile);

  // print out all elements contained in the file
  int result = printSubcircuitPorts(format);
  if(result < 0)
    return result;
  result = printComponents(format);
  if(result < 0)
    return result;
  result = printWires(format);
  if(result < 0)
    return result;

  if(format != outputFormatText)
    fprintf(outputFile,
      "<Diagrams>\n"
      "</Diagrams>\n"
      "<Paintings>\n"
      "</Paintings>\n");

  return 0;
}

// ------------------------------------------------------------------------
int FoaHandler::printSubcircuitPorts(int format)
{
  char *p, *ps, *name;
  int numXY, numPorts, numAngles;

  // get all subcircuit ports
  // The port parameters are not needed here. They are stored just like
  // component parameters, but in block 0x000e with "type1"=0F and "type2"=0D
  p = getBlock(0x0103, &numPorts);
  if(p == 0) 
    return 0; // no ports exist
  char *portName = getSubblock(p, 0x0300);
  if(portName == 0)
    return -101;
  char *portNo = getSubblock(p, 0x0305);
  if(portNo == 0)
    return -102;

  // rotation angle
  double *portAngle = 0;
  p = getBlock(0x0111, &numAngles);
  if(p == 0) 
    return -103;
  p = getSubblock(p, 0x0102);
  if(p) {
    ps = getSubblock(p, 0x0200);
    if(ps == 0)
      return -105;
    portAngle = (double*)getSubblock(p, *(int64_t*)(ps+4) | 0x8000000000000000);
    if(portAngle == 0)
      return -106;
    if(portAngle+numAngles+2 >= (double*)fileEnd)
      return -106;
  }

  // get coordinates
  p = getBlock(0x010F, &numXY);
  if(p == 0)
    return -107;
  char *types = getSubblock(p, 0x0103);
  if(types == 0)
    return -108;
  char *indeces = getSubblock(p, 0x0104);
  if(indeces == 0)
    return -109;
  int32_t *xyPos = (int*)getSubblock(p, 0x0304);
  if(xyPos == 0)
    return -110;
  if(xyPos+4*numXY+4 >= (int32_t*)fileEnd)
    return -110;

  if(format == outputFormatText)
    fprintf(outputFile, "Ports:\n");

  double angle;
  int i, idx, no, x, y, tx, ty;
  for(i=0; i<numXY; i++) {
    if(getValue(types, i) != 0x2062) // not a port?
      continue;
    idx = getValue(indeces, i) - 1;
    if(idx >= numPorts)
      return -111;
    if(idx >= numAngles)
      return -112;

    name = getObjectName(getValue(portName, idx) / 2);
    if(name == 0)
      return -113;
    no = getValue(portNo, idx);
    if(no < 0)
      return -114;
    x = xyPos[4*i+1];      // port coordinates
    y = xyPos[4*i+2];
    tx = xyPos[4*(i+1)+1]; // text coordinates
    ty = xyPos[4*(i+1)+4];
    if(portAngle)
      angle = portAngle[idx+1];
    else
      angle = 0.0;
      
    if(format == outputFormatText)
      fprintf(outputFile,
          "  %s: no=%d, x=%d, y=%d, text=%d/%d, angle=%0.1f\n",
           name, no, x, y, tx, ty, angle);
    else
      fprintf(outputFile,
          "  <Port %s 1 %d %d %d %d 0 %d \"%d\" 1 \"analog\" 0>\n",
                 name, x/2, y/-2, tx/2-x, ty/-2-y, int(angle)/90, no);
  }

  fflush(outputFile);
  return numPorts;
}

// ------------------------------------------------------------------------
int FoaHandler::getComponentParameters(
    int numComponents, int *numParameters, char** &paramValue, char** &paramVisible)
{
  int *pi = (int*)getBlock(0x0007);
  if(pi == 0)
    return -301;
  int n = 3*pi[0] + 20;
  if(pi+n >= (int*)fileEnd)
    return -302;
  n = pi[n];
  if(n == 0) // no component parameters?
    return 0;

  int i, j, count;
  char *p = getBlock(int64_t(n) | 0x8000000000000000, &count);
  if(p == 0)  // no component parameters?
    return 0;
  if(count < 0)  // no parameter "count"?
    return -303;
  char *paramPos = getSubblock(p, 0xFFFFFFFFFFFFFFFF);
  if(paramPos == 0)
    return -304;
  char *parameterType = getSubblock(p, 0x0103);
  if(parameterType == 0)
    return -305;
  p = getSubblock(p, 0x0102); // get parameter subblock
  if(p == 0)
    return -306;
  /* This block does not always exist!
  pi = (int*)getSubblock(p, 0x0204); // get block IDs
  if(pi == 0)
    return -307;
  char *parameterValue = getSubblock(p, int64_t(pi[5]) | 0x8000000000000000);
  if(parameterValue == 0)
    return -308;
  */
  pi = (int*)getSubblock(p, 0x0200); // get block IDs
  if(pi == 0)
    return -307;
  if(pi[0] < 3)
    return -308;
  j = 1;
  for(i=n=0; i<pi[0]; i++) { // search for block ID with parameter values
    if(pi[j+3] == 0x0042) {
      n++;  // for parameters: n=1 are names, n=2 are entry forms, n=3 are values
      if(n == 3)
        break;
      j += 7;
    }
    else if(pi[j+3] == 0x0040)
      j += 6;
    else
      return -309;
  }
  char *parameterValue = getSubblock(p, int64_t(pi[j]) | 0x8000000000000000);
  if(parameterValue == 0)
    return -310;

  // get numbers and positions of component parameters
  p = getBlock(0x000E, &count);
  if(p == 0)
    return -311;
  char *type1 = getSubblock(p, 0x0303);
  if(type1 == 0)
    return -312;
  char *type2 = getSubblock(p, 0x0103);
  if(type2 == 0)
    return -313;
  char *index = getSubblock(p, 0x0304);
  if(index == 0)
    return -314;
  char *paramIndex = getSubblock(p, 0x0302);
  if(paramIndex == 0)
    return -315;

  int **paramStart = (int**)malloc(numComponents * sizeof(int*));
  if(paramStart == 0)
    return -316;

  // index list that contains the component parameters in correct order
  if(getValue(type1, count-1) < 0) // check if "count" make sense
    return -317;
  n = 2*count * sizeof(char*);
  paramValue = (char**)malloc(n);
  if(paramValue == 0) {
    free(paramStart);
    return -318;
  }
  paramVisible = paramValue + count; // parameter visible or hidden
  memset(paramValue, 0, n);

  // go through the list in order to
  // determine the numbers and positions of component parameters
  int k, m, ii, idx, pIdx, count2, *params, *paramPosNew;
  params = paramPosNew = (int*)malloc(count * sizeof(int));
  if(paramPosNew == 0) {
    free(paramStart);
    free(paramValue);
    return -319;
  }
  for(i=0; i<count; i++) {
    if(getValue(type1, i) != 0x13) // not a component?
      continue;
    if(getValue(type2, i) != 0x0D) // not a component?
      continue;

    // n = the index of the component whose parameters are handled now
    // (same values as in block 0x010F, subblock 0x0311, subsubblock 0x0400)
    n = getValue(paramIndex, i);
    if((n < 1) || (n > numComponents))
      break;  // ERROR!
    count2 = 0; // number of parameters
    for(j=0; j<count; j++) {
      if(getValue(type1, j) != 0x13) // not a parameter set?
        continue;
      if(getValue(type2, j) != 0x05) // not a parameter set?
        continue;
      if(getValue(paramIndex, j) != n) // not the wanted parameter set?
        continue;

      idx = getValue(index, j);
      if((idx < 1) || (idx > count))
        break;
      for(k=0; k<count; k++) {
        if(getValue(type1, k) != 0x53) // not a parameter?
          continue;
        ii = getValue(type2, k);
        if(ii != 0x08)   // not a standard parameter?
          if(ii != 0x00) // not a parameter of a variable?
            continue;
        if(getValue(index, k) != idx) // not the wanted parameter?
          continue;
        if(ii == 0) // parameter of a variable?
          params[count2++] = -1;
        params[count2++] = getValue(paramIndex, k);
      }
      break;
    }

    if((count2 > 0) && (params[0] < 0)) { // variable component?
      // A variable component has three parameters per variable:
      // an empty one, the variable name, the variable value.
      // Here, they are convert to two standard ones.
      ii = 0;
      for(j=0; j<count2; j+=2) {
        idx = params[j+1];
        for(k=0; k<count; k++) {
          if(getValue(type1, k) != 0x53) // not a parameter?
            continue;
          if(getValue(type2, k) != 0x0C) // not an equation parameter?
            continue;
          if(getValue(paramIndex, k) != idx) // not the wanted parameter?
            continue;
          idx = getValue(index, k);
          break;
        }
        for(k=0; k<count; k++) {
          if(getValue(type1, k) != 0x53) // not a parameter?
            continue;
          if(getValue(type2, k) != 0x08) // not a parameter?
            continue;
          if(getValue(index, k) != idx) // not the wanted parameter?
            continue;
          params[ii++] = getValue(paramIndex, k);
        }
      }
    }

    n--;
    paramStart[n] = params;
    numParameters[n] = count2;
    params += count2;
  }
  if(i < count) { // error occured?
    free(paramPosNew);
    free(paramStart);
    free(paramValue);
    return -399;
  }


  // create list with component parameters in paramValue[] and paramVisible[]
  pIdx = 0;
  for(i=0; i<numComponents; i++) {
    pi = paramStart[i];

    if(numParameters[i] > 0) {
      // find the first parameter
      // This is really not easy! It seems that the only way for this
      // is to look for the parameter which has no predecessor.
      for(j=0; j<numParameters[i]; j++) { // test all parameters
        for(k=0; k<numParameters[i]; k++) { // look for predecessors
          m = getValue(paramPos, pi[k]-1); // index of next parameter
          if(m == pi[j]) // next index equals current index?
            break;  // parameter[j] has a predecessor
        }
        if(k >= numParameters[i]) // no predecessor found?
          break;
      }
      k = pi[j]; // index of first parameter
    }

    // create list of parameters in paramValue[] and paramVisible[]
    for(j=0; j<numParameters[i]; j++) {
      paramValue[pIdx] = getString(parameterValue, k);
      paramVisible[pIdx] = (char*)(getValue(parameterType, k-1)&0x4000 ? "1" : "0");
      pIdx++;

      k = getValue(paramPos, k-1); // get next position
      if(k < 1)
        if((j+1) < numParameters[i])
          k = pi[j+1];
    }
  }
  // now it is: pIdx = number of used parameters

  free(paramPosNew);
  free(paramStart);
  return pIdx;
}

// ------------------------------------------------------------------------
int FoaHandler::printComponents(int format)
{
  char *p, *flag, *status;
  int i, j, n, ii, r, numComponents;

  // get the component views
  p = getBlock(0x0105);
  if(p == 0)
    return 0;  // file contains no components
  char *modelLib = getSubblock(p, 0x0400); // names of libraries (e.g. "rflib")
  if(modelLib == 0)
    return -201;
  char *modelName = getSubblock(p, 0x0401); // names of models (e.g. "R")
  if(modelName == 0)
    return -202;
//  char *modelView = getSubblock(p, 0x0402); // names of views (e.g. "symbol")
//  char *modelPath = getSubblock(p, 0x0403); // paths (e.g. "rflib.R.symbol")

  // get all components
  p = getBlock(0x0107, &numComponents);
  if(p == 0)
    return -203;
  char *componentName = getSubblock(p, 0x0407); // instance name
  if(componentName == 0)
    return -204;
  char *modelIdx = getSubblock(p, 0x0300);
  if(modelIdx == 0)
    return -205;
  int *componentXY = (int*)getSubblock(p, 0x0301); // coordinates
  if(componentXY == 0)
    return -206;
  status = getSubblock(p, 0x0103);
  if(status == 0)
    return -207;

  // get coordinates of component text
  int count=0;
  p = getBlock(0x010F, &count);
  if(p == 0)
    return -208;
  p = getSubblock(p, 0x0311, &count);
  if(p == 0)
    return -209;
  char *compOrder = getSubblock(p, 0x0400);
  if(compOrder == 0)
    return -210;
  char *coordinateType = getSubblock(p, 0x0402);
  if(coordinateType == 0)
    return -211;
  flag = getSubblock(p, 0x0307); // flag if component or not?
  if(flag == 0)
    return -212;
  int *textCoordinates = (int*)getSubblock(p, 0x0302);
  if(textCoordinates == 0)
    return -213;

  // get inactive status
  char *statusIndex=0, *statusType=0;
  p = getBlock(0x000C, &count);
  if(p) {
    statusIndex = getSubblock(p, 0x0301);
    if(statusIndex == 0)
      return -215;
    statusType = getSubblock(p, 0x0103);
    if(statusType == 0)
      return -216;
  }

  // allocate memory for component properties
  if(getValue(status, numComponents-1) < 0) // check if "numComponents" makes sense
    return -217;
  int *textXY, *componentProperty, *numParameters;
  n = 4*numComponents * sizeof(int);
  componentProperty = (int*)malloc(n);
  if(componentProperty == 0)
    return -218;
  textXY = componentProperty + numComponents;
  numParameters = textXY + 2*numComponents;
  memset(componentProperty, 0, n);

  // get some component properties
  for(i=0; i<numComponents; i++) {
    n = getValue(status, i); // properties like rotate, mirror etc.
    n &= ~3; // make space for inactive bits
    if((n & 0x4000) == 0)
      n |= 1;  // set inactive (not shorted)
    componentProperty[i] = n;
  }

  // get inactive status of component
  if(statusType) {
    for(i=0; i<count; i++) {
      j = getValue(statusType, i);
      if(j < 0)
        break;
      if(j != 0x38)  // component not shorted?
        continue;
      j = getValue(statusIndex, i) - 1;
      if((j < 0) || (j >= numComponents))
        break;
      componentProperty[j] += 2; // set shorted
    }
    if(i < count) {
      free(componentProperty);
      return -219;
    }
  }

  for(i=0; i<count; i++) {
    j = getValue(flag, i);
    if(j < 0) {
      free(componentProperty);
      return -220;
    }
    if(j == 0x1C) // component?
      if(getValue(coordinateType, i) == 2) { // coordinate of component model text?
        j = getValue(compOrder, i) - 1;
        if((j < 0) || (j >= numComponents)) {
          free(componentProperty);
          return -221;
        }
        textXY[2*j] = textCoordinates[1];
        textXY[2*j+1] = textCoordinates[2];
      }
    textCoordinates += 2;
  }

  // get component parameters
  char **paramValue = 0;
  char **paramVisible = 0;
  n = getComponentParameters(numComponents, numParameters, paramValue, paramVisible);
  if(n < 0) {
    free(componentProperty);
    return n;
  }


  if(format == outputFormatText)
    fprintf(outputFile, "Components:\n");

  n = 0; // parameter index
  for(i=0; i<numComponents; i++) {
    if(componentProperty[i] & 0x8000) // component not in use?
      continue;

    ii = getValue(modelIdx, i) - 1;
    char *name = getObjectName(getValue(componentName, i) / 2);
    char *lib = getObjectName(getValue(modelLib, ii)-1);
    char *model = getObjectName(getValue(modelName, ii)-1);
    if((name == 0) || (lib == 0) || (model == 0)) {
      r = -299;
      break;
    }

    if(format == outputFormatText)  // output as text list?
      r = outputTextFormat(outputFile, name, lib, model,
                           componentProperty[i], componentXY+2*i+1, textXY+2*i,
                           numParameters[i], paramValue+n, paramVisible+n);
    else  // output as Qucs schematic
      r = outputQucsFormat(outputFile, name, lib, model,
                           componentProperty[i], componentXY+2*i+1, textXY+2*i,
                           numParameters[i], paramValue+n, paramVisible+n);
    if(r < 0)
      break;
    n += numParameters[i];
  }

  if(format != outputFormatText)
    fprintf(outputFile, "</Components>\n");
  fflush(outputFile);

  free(paramValue);
  free(componentProperty);
  if(r < 0)
    return r;
  return numComponents;
}

// ------------------------------------------------------------------------
int FoaHandler::printWires(int format)
{
  // get coordinates
  int numXY;
  char *p = getBlock(0x010f, &numXY);
  if(p == 0)
    return -401;
  char *types = getSubblock(p, 0x0103);
  if(types == 0)
    return -402;
  char *indeces = getSubblock(p, 0x0104);
  if(indeces == 0)
    return -403;

  // get wires and their labels
  char *labelName=0, *labelIndex=0;
  int numWires, *labelXY=0;
  char *pBlock = getSubblock(p, 0x030e);
  if(pBlock) {
    labelIndex = getSubblock(p, 0x0300);
    if(labelIndex == 0)
      return -405;
    labelName = getSubblock(pBlock, 0x0302); // label name
    if(labelName == 0)
      return -406;
    labelXY = (int*)getSubblock(pBlock, 0x0304); // label coordinates
    if(labelXY == 0)
      return -407;
  }
  char *wireXY = 0;
  pBlock = getSubblock(p, 0x030a, &numWires);
  if(pBlock) {
    wireXY = getSubblock(pBlock, 0x0300);
    if(wireXY == 0)
      return -408;
  }
  if(wireXY == 0) // no wires?
    return 0;


  int i, j, k, n1, n2, x, y, x1, y1, pos, len, idx, type, count;
  n1 = ((uint32_t*)wireXY)[1];  // bytes per value
  if((n1 < 1) || (n1 > 2))
    return -409;
  char *posLen = wireXY + ((n1*numWires+3) & ~3) + 12;
  if(posLen >= fileEnd)
    return -410;
  n2 = *((uint32_t*)(posLen - 4));
  if((n2 < 1) || (n2 > 2))
    return -411;
  char *posXY = posLen + ((n2*numWires+3) & ~3) + 19;
  if(posXY >= fileEnd)
    return -412;

  if(format == outputFormatText)
    fprintf(outputFile, "Wires:\n");
  else
    fprintf(outputFile, "<Wires>\n");

  // output wires
  for(i=0; i<numXY; i++) {
    type = getValue(types, i);
    if(type < 0)
      return -413;
    if((type & 0xFF) != 0x24) // Is it not a wire?
      continue;
    idx = getValue(indeces, i) - 1;
    if((idx < 0) || (idx > numWires))
      return -414;

    pos = *((uint32_t*)(wireXY+idx*n1+8));
    if(n1 == 1)
      pos &= 0x00FF;
    else
      pos &= 0xFFFF;
    if(pos < 1)
      return -415;
    len = *((uint32_t*)(posLen + idx*n2));
    if(n2 == 1)
      len &= 0x00FF;
    else
      len &= 0xFFFF;
    len /= 4;  // number of wire coordinates = number of wire segments + 1

    // Wires are stored as polylines, but here we split them into their segments.
    x = y = x1 = y1 = 0;
    p = posXY + pos;
    for(j=0; j<len; j++) {
      if(p >= fileEnd) // outside of file data?
        return -416;
      x += getWireCoordinate(p);
      if(p >= fileEnd) // outside of file data?
        return -417;
      y += getWireCoordinate(p);

      if(j > 0) {
        int labelX = 0;
        int labelY = 0;
        const char *labelStr = "";
        if(labelIndex && (type == 0x2024)) { // wire with label?
          idx = getValue(labelIndex, i);
          for(k=0; k<numXY; k++) { // search label that belongs to this wire
            type = getValue(types, k);
            if(type == 0x2029) {  // Is it a wire label?
              if(getValue(labelIndex, k) == idx) {  // belongs it to this wire?
                count = getValue(indeces, k);
                labelStr = getString(labelName, count);
                labelX = labelXY[2*count-1];
                labelY = labelXY[2*count];
                break;
              }
            }
          }
        }
        if(format == outputFormatText) {
          fprintf(outputFile, "  from %d,%d to %d,%d", x, y, x1, y1);
          if(labelStr[0] != 0)
            fprintf(outputFile, ", label \"%s\" at %d,%d\n", labelStr, labelX, labelY);
          else
            fprintf(outputFile, "\n");
        }
        else
          fprintf(outputFile, "  <%d %d %d %d \"%s\" %d %d 0 \"\">\n",
                  x/8*3, y/-8*3, x1/8*3, y1/-8*3, labelStr, labelX/8*3, labelY/-8*3-TEXT_SIZE-5);
      }
      x1 = x;
      y1 = y;
    }
  }

  if(format != outputFormatText)
    fprintf(outputFile, "</Wires>\n");

  fflush(outputFile);
  return 1;
}

// ------------------------------------------------------------------------
// ----------                    Main routine                    ----------
// ------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  if(argc < 2) { // no arguments given in the command line?
    printf("\n"
           "OpenAccess file converter (version 1.0)\n\n"
           "Usage: %s [-hex|-txt|-qucs] input_filename.oa [further_names.oa]\n\n"
           "Creates output file(s) \"input_filename.hex|txt|sch\" which "
           "contains the decoded content of the input file(s).\n\n"
           "The option specifies the output format:\n"
           "    \"-txt\"  - (the default) text file with elements (components, wires etc.)\n"
           "    \"-hex\"  - text file with blocks as hexadecimal number arrays\n"
           "    \"-qucs\" - Qucs schematic\n\n",
           argv[0]);
    return 1;
  }

  int i = 1;
  int outputFormat = FoaHandler::outputFormatText; // -txt
  const char *suffix = ".txt";
  if(argv[1][0] == '-') {
    if(strcmp(argv[1], "-hex") == 0) {
      outputFormat = FoaHandler::outputFormatHex;
      suffix = ".hex";
    }
    else if(strcmp(argv[1], "-qucs") == 0) {
      outputFormat = FoaHandler::outputFormatQucs;
      suffix = ".sch";
    }
    else if(strcmp(argv[1], "-txt") != 0)
      printf("WARNING: Unknown output format (%s). Fall back to default.\n", argv[1]);
    i++;
  }

  FoaHandler oaFile;
  for(; i<argc; i++) {
    char *filename = argv[i];
    printf("\nreading file: %s\n", filename);
    fflush(stdout);

    // read all file data into memory
    if(oaFile.loadFile(filename) < 0) {
      fprintf(stderr, "ERROR: Cannot read this file!\n");
      return -1;
    }
    if(oaFile.checkFileFormat() < 0) {
      fprintf(stderr, "ERROR: This is not an OpenAccess file!\n");
      continue;//return -2;  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    }

    // output filename = input filename with suffix ".oa" replaced with ".txt"
    if(oaFile.openOutputFile(filename, suffix) < 0) {
      fprintf(stderr, "ERROR: Cannot write this file!\n");
      return -3;
    }

    // convert data and output result
    int result = -1;
    if(outputFormat == FoaHandler::outputFormatHex)
      result = oaFile.printFileContent(); // as hexadecimal & ASCII text
    else // as elements
      result = oaFile.printFileElements(outputFormat);
    oaFile.closeOutputFile();

    if(result < 0) {
      fprintf(stderr, "ERROR #%d: This file is corrupt.\n", -result);
//      return -4; <<<<<<<<<<<<<<<<<<<<<<<<
    } else // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    printf("Finished successfully!\n");
  }

  return 0;
}
